# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import re
import socket
import struct

from scciclient.irmc import scci
from scciclient.irmc.viom import elcm


ONBOARD = 'onboard'
ADDON = 'addon'
_CARD_DICT = {
    ONBOARD: elcm.OnboardCard,
    ADDON: elcm.AddOnCard}

LAN = 'LAN'
FC = 'FC'
CNA = 'CNA'
_ADAPTER_DICT = {
    LAN: elcm.LANAdapter,
    FC: elcm.FCAdapter,
    CNA: elcm.CNAAdapter
}
_POSSIBLE_CARD_TYPE = _ADAPTER_DICT.keys()

_CNA_LAN_FUNC_IDX = 1
_CNA_FCOE_FUNC_IDX = 2
_CNA_ISCSI_FUNC_IDX = 3


class _PortHandler(object, metaclass=abc.ABCMeta):
    """VIOM Configurator with physical port information"""

    def __init__(self, slot_type, card_type, slot_idx, card_idx, port_idx):
        self.slot_type = slot_type
        self.card_type = card_type
        self.slot_idx = slot_idx
        self.card_idx = card_idx
        self.port_idx = port_idx

    def need_padding(self):
        return False

    def create_card(self):
        return _CARD_DICT[self.slot_type](
            self.card_idx, _ADAPTER_DICT[self.card_type]())

    def create_lan_port(self, mac=None, port_enable=True):
        raise NotImplementedError()

    def set_lan_port(self, port, mac=None):
        raise NotImplementedError()

    def set_iscsi_port(self, port, iscsi_boot):
        raise NotImplementedError()

    def create_iscsi_port(self, iscsi_boot):
        raise NotImplementedError()

    def set_fc_port(self, port, fc_boot, wwnn=None, wwpn=None):
        raise NotImplementedError()

    def create_fc_port(self, fc_boot, wwnn=None, wwpn=None):
        raise NotImplementedError()


class _LANPortHandler(_PortHandler):
    """Configurator for LAN card."""

    def need_padding(self):
        return True if self.slot_type == ONBOARD else False

    def create_lan_port(self, mac=None, port_enable=True):
        return elcm.LANPort(self.port_idx,
                            port_enable=port_enable,
                            use_virtual_addresses=bool(mac),
                            mac=mac)

    def set_lan_port(self, port, mac=None):
        port.port_enable = True
        port.mac = mac
        port.use_virtual_addresses = bool(mac)

    def set_iscsi_port(self, port, iscsi_boot):
        port.boot = iscsi_boot

    def create_iscsi_port(self, iscsi_boot):
        return elcm.LANPort(self.port_idx, boot=iscsi_boot, port_enable=True)


class _FCPortHandler(_PortHandler):
    """Configurator for FC card."""

    def set_fc_port(self, port, fc_boot, wwnn=None, wwpn=None):
        port.boot = fc_boot
        port.port_enable = True
        port.use_virtual_addresses = bool(wwnn or wwpn)
        port.wwnn = wwnn
        port.wwpn = wwpn

    def create_fc_port(self, fc_boot, wwnn=None, wwpn=None):
        return elcm.FCPort(self.port_idx, boot=fc_boot, wwnn=wwnn, wwpn=wwpn,
                           use_virtual_addresses=bool(wwnn or wwpn),
                           port_enable=True)


class _CNAPortHandler(_PortHandler):
    """Configurator for CNA card."""

    def _create_port(self, function):
        cna_port = elcm.CNAPort(self.port_idx)
        if not isinstance(function, elcm.LANFunction):
            # LanFunction is must
            cna_port.add_function(
                elcm.LANFunction(_CNA_LAN_FUNC_IDX, function_enable=False,
                                 boot=None))
        cna_port.add_function(function)
        return cna_port

    def create_lan_port(self, mac=None, port_enable=True):
        function = elcm.LANFunction(_CNA_LAN_FUNC_IDX,
                                    function_enable=port_enable,
                                    use_virtual_addresses=bool(mac),
                                    mac=mac)
        return self._create_port(function)

    def set_lan_port(self, port, mac=None):
        function = port.get_function(_CNA_LAN_FUNC_IDX)
        # Lan Function is always created when port is created.
        function.function_enable = True
        function.mac = mac
        function.use_virtual_addresses = bool(mac)

    def set_iscsi_port(self, port, iscsi_boot):
        function = port.get_function(_CNA_ISCSI_FUNC_IDX)
        if function:
            function.boot = iscsi_boot
        else:
            function = elcm.ISCSIFunction(_CNA_ISCSI_FUNC_IDX, boot=iscsi_boot,
                                          function_enable=True)
            port.add_function(function)

    def create_iscsi_port(self, iscsi_boot):
        function = elcm.ISCSIFunction(_CNA_ISCSI_FUNC_IDX, boot=iscsi_boot)
        return self._create_port(function)

    def set_fc_port(self, port, fc_boot, wwnn=None, wwpn=None):
        function = port.get_function(_CNA_FCOE_FUNC_IDX)
        if function:
            function.boot = fc_boot
            function.use_virtual_addresses = bool(wwnn or wwpn)
            function.wwnn = wwnn
            function.wwpn = wwpn
        else:
            function = elcm.FCoEFunction(
                _CNA_FCOE_FUNC_IDX, boot=fc_boot, function_enable=True,
                use_virtual_addresses=bool(wwnn or wwpn), wwnn=wwnn, wwpn=wwpn)
            port.add_function(function)

    def create_fc_port(self, fc_boot, wwnn=None, wwpn=None):
        function = elcm.FCoEFunction(
            _CNA_FCOE_FUNC_IDX, boot=fc_boot, function_enable=True,
            use_virtual_addresses=bool(wwnn or wwpn), wwnn=wwnn, wwpn=wwpn)
        return self._create_port(function)


_PORT_HANDLERS = {
    LAN: _LANPortHandler,
    FC: _FCPortHandler,
    CNA: _CNAPortHandler,
}


def _parse_physical_port_id(port_id):

    message = ('Physical port ID should follow the format: '
               '<card-type><slot-idx>-<port-idx> like CNA1-1. '
               '<card-type> must be chosen from CNA, FC, or LAN. '
               '<slot-idx> should be 0 for onboard slot or 1-9 for addon '
               'slot. <port-idx> should be 1-9.')

    m = re.match('^([a-zA-Z]+)([0-9])-([1-9])$', port_id)
    if not m:
        raise scci.SCCIInvalidInputError(message)

    card_type = m.group(1).upper()
    if card_type not in _POSSIBLE_CARD_TYPE:
        raise scci.SCCIInvalidInputError(message)

    slot_idx = 0
    if int(m.group(2)) == 0:
        slot_type = ONBOARD
        card_idx = 1
    else:
        slot_type = ADDON
        card_idx = int(m.group(2))
    port_idx = int(m.group(3))

    return _PORT_HANDLERS[card_type](slot_type, card_type, slot_idx, card_idx,
                                     port_idx)


def _create_iscsi_boot(initiator_iqn,
                       initiator_dhcp=False, initiator_ip=None,
                       initiator_netmask=None,
                       target_dhcp=False, target_iqn=None, target_ip=None,
                       target_port=None, target_lun=None, boot_prio=1,
                       chap_user=None, chap_secret=None,
                       mutual_chap_secret=None):
    iscsi_initiator = elcm.ISCSIInitiator(dhcp_usage=initiator_dhcp,
                                          iqn=initiator_iqn,
                                          ip=initiator_ip,
                                          subnet=initiator_netmask)
    if chap_user and chap_secret:
        auth_method = 'MutualCHAP' if mutual_chap_secret else 'CHAP'
    else:
        auth_method = 'None'
    iscsi_target = elcm.ISCSITarget(dhcp_usage=target_dhcp,
                                    iqn=target_iqn,
                                    ip=target_ip,
                                    port=target_port,
                                    lun=target_lun,
                                    auth_method=auth_method,
                                    chap_user=chap_user,
                                    chap_secret=chap_secret,
                                    mutual_chap_secret=mutual_chap_secret)
    iscsi_boot = elcm.ISCSIBoot(iscsi_initiator,
                                iscsi_target,
                                boot_prio=boot_prio)
    return iscsi_boot


def _convert_netmask(mask):
    """Convert netmask from CIDR format(integer) to doted decimal string."""
    if mask not in range(0, 33):
        raise scci.SCCIInvalidInputError(
            'Netmask value is invalid.')

    return socket.inet_ntoa(struct.pack(
        '!L', int('1' * mask + '0' * (32 - mask), 2)))


class VIOMConfiguration(object):
    """VIOM Configurator for volume boot"""

    def __init__(self, irmc_info, identification):
        self.client = elcm.ELCMVIOMClient(irmc_info)
        self.root = elcm.VIOMTable()
        self.root.set_manage_table(
            elcm.ManageTable(identification=identification))

    def apply(self, reboot=False):
        """Apply the configuration to iRMC."""
        self.root.use_virtual_addresses = True
        self.root.manage.manage = True
        self.root.mode = 'new'
        self.root.init_boot = reboot

        self.client.set_profile(self.root.get_json())

    def terminate(self, reboot=False):
        """Delete VIOM configuration from iRMC."""
        self.root.manage.manage = False
        self.root.mode = 'delete'
        self.root.init_boot = reboot
        self.client.set_profile(self.root.get_json())

    def _find_card(self, port_handler):
        slot = self.root.get_slot(port_handler.slot_idx, create=False)
        if not slot:
            return None
        if port_handler.slot_type == ONBOARD:
            return slot.get_onboard_card(port_handler.card_idx)
        else:
            return slot.get_addon_card(port_handler.card_idx)

    def _get_card(self, port_handler):
        card = self._find_card(port_handler)
        if card:
            return card
        card = port_handler.create_card()
        self.root.get_slot(port_handler.slot_idx).add_card(card)
        return card

    def _find_port(self, port_handler):
        card = self._find_card(port_handler)
        if not card:
            return None
        return card.get_port(port_handler.port_idx)

    def _add_port(self, port_handler, port):
        self._pad_former_ports(port_handler)
        card = self._get_card(port_handler)
        card.add_port(port)

    def set_lan_port(self, port_id, mac=None):
        """Set LAN port information to configuration.

        :param port_id: Physical port ID.
        :param mac: virtual MAC address if virtualization is necessary.
        """
        port_handler = _parse_physical_port_id(port_id)
        port = self._find_port(port_handler)
        if port:
            port_handler.set_lan_port(port, mac)
        else:
            self._add_port(port_handler, port_handler.create_lan_port(mac))

    def set_iscsi_volume(self, port_id,
                         initiator_iqn, initiator_dhcp=False,
                         initiator_ip=None, initiator_netmask=None,
                         target_dhcp=False, target_iqn=None, target_ip=None,
                         target_port=3260, target_lun=0, boot_prio=1,
                         chap_user=None, chap_secret=None,
                         mutual_chap_secret=None):
        """Set iSCSI volume information to configuration.

        :param port_id: Physical port ID.
        :param initiator_iqn: IQN of initiator.
        :param initiator_dhcp: True if DHCP is used in the iSCSI network.
        :param initiator_ip: IP address of initiator. None if DHCP is used.
        :param initiator_netmask: Netmask of initiator as integer. None if
               DHCP is used.
        :param target_dhcp: True if DHCP is used for iSCSI target.
        :param target_iqn: IQN of target. None if DHCP is used.
        :param target_ip: IP address of target. None if DHCP is used.
        :param target_port: Port number of target.  None if DHCP is used.
        :param target_lun: LUN number of target. None if DHCP is used,
        :param boot_prio: Boot priority of the volume. 1 indicates the highest
            priority.
        """

        initiator_netmask = (_convert_netmask(initiator_netmask)
                             if initiator_netmask else None)

        port_handler = _parse_physical_port_id(port_id)
        iscsi_boot = _create_iscsi_boot(
            initiator_iqn,
            initiator_dhcp=initiator_dhcp,
            initiator_ip=initiator_ip,
            initiator_netmask=initiator_netmask,
            target_dhcp=target_dhcp,
            target_iqn=target_iqn,
            target_ip=target_ip,
            target_port=target_port,
            target_lun=target_lun,
            boot_prio=boot_prio,
            chap_user=chap_user,
            chap_secret=chap_secret,
            mutual_chap_secret=mutual_chap_secret)

        port = self._find_port(port_handler)
        if port:
            port_handler.set_iscsi_port(port, iscsi_boot)
        else:
            port = port_handler.create_iscsi_port(iscsi_boot)
            self._add_port(port_handler, port)

    def set_fc_volume(self, port_id,
                      target_wwn, target_lun=0, boot_prio=1,
                      initiator_wwnn=None, initiator_wwpn=None):
        """Set FibreChannel volume information to configuration.

        :param port_id: Physical port ID.
        :param target_wwn: WWN of target.
        :param target_lun: LUN number of target.
        :param boot_prio: Boot priority of the volume. 1 indicates the highest
               priority.
        :param initiator_wwnn: Virtual WWNN for initiator if necessary.
        :param initiator_wwpn: Virtual WWPN for initiator if necessary.
        """
        port_handler = _parse_physical_port_id(port_id)
        fc_target = elcm.FCTarget(target_wwn, target_lun)
        fc_boot = elcm.FCBoot(boot_prio=boot_prio, boot_enable=True)
        fc_boot.add_target(fc_target)

        port = self._find_port(port_handler)
        if port:
            port_handler.set_fc_port(port, fc_boot,
                                     wwnn=initiator_wwnn, wwpn=initiator_wwpn)
        else:
            port = port_handler.create_fc_port(fc_boot,
                                               wwnn=initiator_wwnn,
                                               wwpn=initiator_wwpn)
            self._add_port(port_handler, port)

    def dump_json(self):
        """Create JSON profile based on current configuration.

        :returns: JSON data created from current configurtion. It can be
                  logged by a caller.
        """
        return self.root.get_json()

    def _pad_former_ports(self, port_handler):
        """Create ports with former port index.

        :param port_handler: Port information to be registered.

        Depending on slot type and card type, it is necessary to register
        LAN ports with former index to VIOM table.
        """
        if not port_handler.need_padding():
            return
        for port_idx in range(1, port_handler.port_idx):
            pad_handler = port_handler.__class__(
                port_handler.slot_type,
                port_handler.card_type,
                port_handler.slot_idx,
                port_handler.card_idx,
                port_idx)
            if not self._find_port(pad_handler):
                self._add_port(pad_handler,
                               pad_handler.create_lan_port())


def validate_physical_port_id(port_id):
    """Validate physical port ID.

    Physical port ID is required for configure interfaces with VIOM API. The
    format is:
        <Card type><Slot Idx>-<Port Idx>

    * Card type is chosen from CNA, FC or LAN.
    * Slot Idx should be 0, which indecates on-board slot or 1-9, which
      specify add-on slot index.
    * Port Idx should be 1-9, which indicate port number.

    :param port_id: Physical port ID following the format.
    """
    _parse_physical_port_id(port_id)



# Copyright 2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
eLCM functionality.
"""

import collections
import time

from oslo_serialization import jsonutils
import requests

from scciclient.irmc import scci

"""
List of profile names
"""
PROFILE_BIOS_CONFIG = 'BiosConfig'
PROFILE_RAID_CONFIG = 'RAIDAdapter'

"""
List of URL paths for profiles
"""
URL_PATH_PROFILE_MGMT = '/rest/v1/Oem/eLCM/ProfileManagement/'

"""
List of request params for profiles
"""
PARAM_PATH_SYSTEM_CONFIG = 'Server/SystemConfig/'
PARAM_PATH_BIOS_CONFIG = PARAM_PATH_SYSTEM_CONFIG + PROFILE_BIOS_CONFIG

PARAM_PATH_HW_CONFIG = 'Server/HWConfigurationIrmc/Adapters/'
PARAM_PATH_RAID_CONFIG = PARAM_PATH_HW_CONFIG + PROFILE_RAID_CONFIG

"""
Timeout values
"""
PROFILE_CREATE_TIMEOUT = 300  # 300 secs
PROFILE_SET_TIMEOUT = 300  # 300 secs
BIOS_CONFIG_SESSION_TIMEOUT = 30 * 60  # 30 mins
RAID_CONFIG_SESSION_TIMEOUT = 30 * 60  # 30 mins

BIOS_CONFIGURATION_DICTIONARY = {
    "boot_option_filter": "CsmConfig_BootOptionFilter",
    "boot_removable_media_enabled": "BootConfig_BootRemovableMediaEnabled",
    "check_controllers_health_status_enabled":
        "BootConfig_CheckControllersHealthStatusEnabled",
    "cpu_active_processor_cores": "CpuConfig_ActiveProcessorCores",
    "cpu_adjacent_cache_line_prefetch_enabled":
        "CpuConfig_AdjacentCacheLinePrefetchEnabled",
    "cpu_dcu_ip_prefetch_enabled": "CpuConfig_DcuIpPrefetchEnabled",
    "cpu_early_snoop": "CpuConfig_EarlySnoop",
    "cpu_energy_performance_mode": "CpuConfig_EnergyPerformanceMode",
    "cpu_enhanced_speed_step_enabled": "CpuConfig_EnhancedSpeedStepEnabled",
    "cpu_execute_disable_bit_enabled": "CpuConfig_ExecuteDisableBitEnabled",
    "cpu_frequency_floor_overwrite_enabled":
        "CpuConfig_FrequencyFloorOverwriteEnabled",
    "cpu_hardware_prefetcher_enabled": "CpuConfig_HardwarePrefetcherEnabled",
    "cpu_power_technology": "CpuConfig_PowerTechnology",
    "cpu_turbo_mode_enabled": "CpuConfig_TurboModeEnabled",
    "cpu_uncore_frequencey_override_enabled":
        "CpuConfig_UncoreFrequenceyOverrideEnabled",
    "cpu_vt_enabled": "CpuConfig_VtEnabled",
    "flash_write_enabled": "SecurityConfig_FlashWriteEnabled",
    "hyper_threading_enabled": "CpuConfig_HyperThreadingEnabled",
    "keep_void_boot_options_enabled": "BootConfig_KeepVoidBootOptionsEnabled",
    "launch_csm_enabled": "CsmConfig_LaunchCsmEnabled",
    "limit_cpu_id_maximum_enabled": "CpuConfig_LimitCpuIdMaximumEnabled",
    "memory_mode": "MemoryConfig_MemoryMode",
    "network_stack_enabled": "NetworkStackConfig_Enabled",
    "os_energy_performance_override_enabled":
        "CpuConfig_OsEnergyPerformanceOverrideEnabled",
    "pci_aspm_support": "PciConfig_ASPMSupport",
    "pci_above_4g_decoding_enabled": "PciConfig_Above4GDecodingEnabled",
    "pending_tpm_operation": "TpmConfig_PendingTpmOperation",
    "power_on_source": "PowerConfig_PowerOnSource",
    "power_wake_on_lan_boot": "PowerConfig_WakeOnLanBoot",
    "pxe_boot_option_retry": "BootConfig_PxeBootOptionRetry",
    "pxe_option_rom_policy": "CsmConfig_PxeOptionRomPolicy",
    "quiet_boot_enabled": "BootConfig_QuietBootEnabled",
    "sas_sata_driver": "OnboardDeviceConfig_SasSataDriver",
    "sas_sata_enabled": "OnboardDeviceConfig_SasSataEnabled",
    "sas_sata_option_rom_enabled":
        "OnboardDeviceConfig_SasSataOptionRomEnabled",
    "sata_controller_enabled": "SataConfig_SataControllerEnabled",
    "sata_mode": "SataConfig_SataMode",
    "secure_boot_control_enabled": "SecurityConfig_SecureBootControlEnabled",
    "secure_boot_mode": "SecurityConfig_SecureBootMode",
    "single_root_io_virtualization_support_enabled":
        "PciConfig_SingleRootIOVirtualizationSupportEnabled",
    "storage_option_rom_policy": "CsmConfig_StorageOptionRomPolicy",
    "tpm_hash_policy": "TpmConfig_HashPolicy",
    "tpm_state_enabled": "TpmConfig_TpmStateEnabled",
    "usb_legacy_support": "UsbConfig_LegacySupport",
    "usb_port_disable": "UsbConfig_PortDisable",
    "usb_xhci_mode": "UsbConfig_XHCIMode",
    "video_option_rom_policy": "CsmConfig_VideoOptionRomPolicy"
}


class ELCMInvalidResponse(scci.SCCIError):
    def __init__(self, message):
        super(ELCMInvalidResponse, self).__init__(message)


class ELCMProfileNotFound(scci.SCCIError):
    def __init__(self, message):
        super(ELCMProfileNotFound, self).__init__(message)


class ELCMSessionNotFound(scci.SCCIError):
    def __init__(self, message):
        super(ELCMSessionNotFound, self).__init__(message)


class ELCMSessionTimeout(scci.SCCIError):
    def __init__(self, message):
        super(ELCMSessionTimeout, self).__init__(message)


class SecureBootConfigNotFound(scci.SCCIError):
    def __init__(self, message):
        super(SecureBootConfigNotFound, self).__init__(message)


class ELCMValueError(scci.SCCIError):
    def __init__(self, message):
        super(ELCMValueError, self).__init__(message)


class BiosConfigNotFound(scci.SCCIError):
    def __init__(self, message):
        super(BiosConfigNotFound, self).__init__(message)


def _parse_elcm_response_body_as_json(response):
    """parse eLCM response body as json data

    eLCM response should be in form of:
    _
    Key1: value1  <-- optional -->
    Key2: value2  <-- optional -->
    KeyN: valueN  <-- optional -->

    - CRLF -

    JSON string
    -

    :param response: eLCM response
    :return: json object if success
    :raise ELCMInvalidResponse: if the response does not contain valid
        json data.
    """
    try:
        body = response.text
        body_parts = body.split('\r\n')
        if len(body_parts) > 0:
            return jsonutils.loads(body_parts[-1])
        else:
            return None
    except (TypeError, ValueError):
        raise ELCMInvalidResponse('eLCM response does not contain valid json '
                                  'data. Response is "%s".' % body)


def elcm_request(irmc_info, method, path, **kwargs):
    """send an eLCM request to the server

    :param irmc_info: dict of iRMC params to access the server node
        {
          'irmc_address': host,
          'irmc_username': user_id,
          'irmc_password': password,
          'irmc_port': 80 or 443, default is 443,
          'irmc_auth_method': 'basic' or 'digest', default is 'basic',
          'irmc_client_timeout': timeout, default is 60,
          'irmc_verify_ca': Either a boolean, in which case it controls
                            whether we verify the server's TLS certificate,
                            or a string, in which case it must be a path to
                            a CA bundle to use. Defaults to ``False``.
          ...
        }
    :param method: request method such as 'GET', 'POST'
    :param path: url path for eLCM request
    :returns: requests.Response from SCCI server
    :raises SCCIInvalidInputError: if port and/or auth_method params
             are invalid
    :raises SCCIClientError: if SCCI failed
    """
    host = irmc_info['irmc_address']
    port = irmc_info.get('irmc_port', 443)
    auth_method = irmc_info.get('irmc_auth_method', 'basic')
    userid = irmc_info['irmc_username']
    password = irmc_info['irmc_password']
    client_timeout = irmc_info.get('irmc_client_timeout', 60)
    verify = irmc_info.get('irmc_verify_ca', False)

    # Request headers, params, and data
    headers = kwargs.get('headers', {'Accept': 'application/json'})
    params = kwargs.get('params')
    data = kwargs.get('data')

    auth_obj = None
    try:
        protocol = {80: 'http', 443: 'https'}[port]
        auth_obj = {
            'basic': requests.auth.HTTPBasicAuth(userid, password),
            'digest': requests.auth.HTTPDigestAuth(userid, password)
        }[auth_method.lower()]

    except KeyError:
        raise scci.SCCIInvalidInputError(
            ("Invalid port %(port)d or " +
             "auth_method for method %(auth_method)s") %
            {'port': port, 'auth_method': auth_method})

    try:
        r = requests.request(method,
                             protocol + '://' + host + path,
                             headers=headers,
                             params=params,
                             data=data,
                             verify=verify,
                             timeout=client_timeout,
                             allow_redirects=False,
                             auth=auth_obj)
    except requests.exceptions.RequestException as requests_exception:
        raise scci.SCCIClientError(requests_exception)

    # Process status_code 401
    if r.status_code == 401:
        raise scci.SCCIClientError('UNAUTHORIZED')

    return r


def elcm_profile_get_versions(irmc_info):
    """send an eLCM request to get profile versions

    :param irmc_info: node info
    :returns: dict object of profiles if succeed
        {
            "Server":{
                "@Version": "1.01",
                "AdapterConfigIrmc":{
                    "@Version": "1.00"
                },
                "HWConfigurationIrmc":{
                    "@Version": "1.00"
                },
                "SystemConfig":{
                    "IrmcConfig":{
                        "@Version": "1.02"
                    },
                    "BiosConfig":{
                        "@Version": "1.02"
                    }
                }
            }
        }
    :raises: SCCIClientError if SCCI failed
    """
    # Send GET request to the server
    resp = elcm_request(irmc_info,
                        method='GET',
                        path=URL_PATH_PROFILE_MGMT + 'version')

    if resp.status_code == 200:
        return _parse_elcm_response_body_as_json(resp)
    else:
        raise scci.SCCIClientError(('Failed to get profile versions with '
                                    'error code %s' % resp.status_code))


def elcm_profile_list(irmc_info):
    """send an eLCM request to list all profiles

    :param irmc_info: node info
    :returns: dict object of profiles if succeed
        {
          'Links':
          {
            'profileStore':
            [
              { '@odata.id': id1 },
              { '@odata.id': id2 },
              { '@odata.id': idN },
            ]
          }
        }
    :raises: SCCIClientError if SCCI failed
    """
    # Send GET request to the server
    resp = elcm_request(irmc_info,
                        method='GET',
                        path=URL_PATH_PROFILE_MGMT)

    if resp.status_code == 200:
        return _parse_elcm_response_body_as_json(resp)
    else:
        raise scci.SCCIClientError(('Failed to list profiles with '
                                    'error code %s' % resp.status_code))


def elcm_profile_get(irmc_info, profile_name):
    """send an eLCM request to get profile data

    :param irmc_info: node info
    :param profile_name: name of profile
    :returns: dict object of profile data if succeed
    :raises: ELCMProfileNotFound if profile does not exist
    :raises: SCCIClientError if SCCI failed
    """
    # Send GET request to the server
    resp = elcm_request(irmc_info,
                        method='GET',
                        path=URL_PATH_PROFILE_MGMT + profile_name)

    if resp.status_code == 200:
        return _parse_elcm_response_body_as_json(resp)
    elif resp.status_code == 404:
        raise ELCMProfileNotFound('Profile "%s" not found '
                                  'in the profile store.' % profile_name)
    else:
        raise scci.SCCIClientError(('Failed to get profile "%(profile)s" with '
                                    'error code %(error)s' %
                                    {'profile': profile_name,
                                     'error': resp.status_code}))


def elcm_profile_create(irmc_info, param_path):
    """send an eLCM request to create profile

    To create a profile, a new session is spawned with status 'running'.
    When profile is created completely, the session ends.

    :param irmc_info: node info
    :param param_path: path of profile
    :returns: dict object of session info if succeed
        {
          'Session':
          {
            'Id': id
            'Status': 'activated'
            ...
          }
        }
    :raises: SCCIClientError if SCCI failed
    """
    # Send POST request to the server
    # NOTE: This task may take time, so set a timeout
    _irmc_info = dict(irmc_info)
    _irmc_info['irmc_client_timeout'] = PROFILE_CREATE_TIMEOUT

    resp = elcm_request(_irmc_info,
                        method='POST',
                        path=URL_PATH_PROFILE_MGMT + 'get',
                        params={'PARAM_PATH': param_path})

    if resp.status_code == 202:
        return _parse_elcm_response_body_as_json(resp)
    else:
        raise scci.SCCIClientError(('Failed to create profile for path '
                                    '"%(param_path)s" with error code '
                                    '%(error)s' %
                                    {'param_path': param_path,
                                     'error': resp.status_code}))


def elcm_profile_set(irmc_info, input_data):
    """send an eLCM request to set param values

    To apply param values, a new session is spawned with status 'running'.
    When values are applied or error, the session ends.

    :param irmc_info: node info
    :param input_data: param values to apply, eg.
        {
          'Server':
          {
            'SystemConfig':
            {
              'BiosConfig':
              {
                '@Processing': 'execute',
                -- config data --
              }
            }
          }
        }
    :returns: dict object of session info if succeed
        {
          'Session':
          {
            'Id': id
            'Status': 'activated'
            ...
          }
        }
    :raises: SCCIClientError if SCCI failed
    """
    # Prepare the data to apply
    if isinstance(input_data, dict):
        data = jsonutils.dumps(input_data)
    else:
        data = input_data

    # Send POST request to the server
    # NOTE: This task may take time, so set a timeout
    _irmc_info = dict(irmc_info)
    _irmc_info['irmc_client_timeout'] = PROFILE_SET_TIMEOUT

    content_type = 'application/x-www-form-urlencoded'
    if input_data['Server'].get('HWConfigurationIrmc'):
        content_type = 'application/json'
    resp = elcm_request(_irmc_info,
                        method='POST',
                        path=URL_PATH_PROFILE_MGMT + 'set',
                        headers={'Content-type': content_type},
                        data=data)

    if resp.status_code == 202:
        return _parse_elcm_response_body_as_json(resp)
    else:
        raise scci.SCCIClientError(('Failed to apply param values with '
                                    'error code %(error)s' %
                                    {'error': resp.status_code}))


def elcm_profile_delete(irmc_info, profile_name):
    """send an eLCM request to delete a profile

    :param irmc_info: node info
    :param profile_name: name of profile
    :raises: ELCMProfileNotFound if the profile does not exist
    :raises: SCCIClientError if SCCI failed
    """
    # Send DELETE request to the server
    resp = elcm_request(irmc_info,
                        method='DELETE',
                        path=URL_PATH_PROFILE_MGMT + profile_name)

    if resp.status_code == 200:
        # Profile deleted
        return
    elif resp.status_code == 404:
        # Profile not found
        raise ELCMProfileNotFound('Profile "%s" not found '
                                  'in the profile store.' % profile_name)
    else:
        raise scci.SCCIClientError(('Failed to delete profile "%(profile)s" '
                                    'with error code %(error)s' %
                                    {'profile': profile_name,
                                     'error': resp.status_code}))


def elcm_session_list(irmc_info):
    """send an eLCM request to list all sessions

    :param irmc_info: node info
    :returns: dict object of sessions if succeed
        {
          'SessionList':
          {
            'Contains':
            [
              { 'Id': id1, 'Name': name1 },
              { 'Id': id2, 'Name': name2 },
              { 'Id': idN, 'Name': nameN },
            ]
          }
        }
    :raises: SCCIClientError if SCCI failed
    """
    # Send GET request to the server
    resp = elcm_request(irmc_info,
                        method='GET',
                        path='/sessionInformation/')

    if resp.status_code == 200:
        return _parse_elcm_response_body_as_json(resp)
    else:
        raise scci.SCCIClientError(('Failed to list sessions with '
                                    'error code %s' % resp.status_code))


def elcm_session_get_status(irmc_info, session_id):
    """send an eLCM request to get session status

    :param irmc_info: node info
    :param session_id: session id
    :returns: dict object of session info if succeed
        {
          'Session':
          {
            'Id': id
            'Status': status
            ...
          }
        }
    :raises: ELCMSessionNotFound if the session does not exist
    :raises: SCCIClientError if SCCI failed
    """
    # Send GET request to the server
    resp = elcm_request(irmc_info,
                        method='GET',
                        path='/sessionInformation/%s/status' % session_id)

    if resp.status_code == 200:
        return _parse_elcm_response_body_as_json(resp)
    elif resp.status_code == 404:
        raise ELCMSessionNotFound('Session "%s" does not exist' % session_id)
    else:
        raise scci.SCCIClientError(('Failed to get status of session '
                                    '"%(session)s" with error code %(error)s' %
                                    {'session': session_id,
                                     'error': resp.status_code}))


def elcm_session_get_log(irmc_info, session_id):
    """send an eLCM request to get session log

    :param irmc_info: node info
    :param session_id: session id
    :returns: dict object of session log if succeed
        {
          'Session':
          {
            'Id': id
            ...
          }
        }
    :raises: ELCMSessionNotFound if the session does not exist
    :raises: SCCIClientError if SCCI failed
    """
    # Send GET request to the server
    resp = elcm_request(irmc_info,
                        method='GET',
                        path='/sessionInformation/%s/log' % session_id)

    if resp.status_code == 200:
        return _parse_elcm_response_body_as_json(resp)
    elif resp.status_code == 404:
        raise ELCMSessionNotFound('Session "%s" does not exist' % session_id)
    else:
        raise scci.SCCIClientError(('Failed to get log of session '
                                    '"%(session)s" with error code %(error)s' %
                                    {'session': session_id,
                                     'error': resp.status_code}))


def elcm_session_terminate(irmc_info, session_id):
    """send an eLCM request to terminate a session

    :param irmc_info: node info
    :param session_id: session id
    :raises: ELCMSessionNotFound if the session does not exist
    :raises: SCCIClientError if SCCI failed
    """
    # Send DELETE request to the server
    resp = elcm_request(irmc_info,
                        method='DELETE',
                        path='/sessionInformation/%s/terminate' % session_id)

    if resp.status_code == 200:
        return
    elif resp.status_code == 404:
        raise ELCMSessionNotFound('Session "%s" does not exist' % session_id)
    else:
        raise scci.SCCIClientError(('Failed to terminate session '
                                    '"%(session)s" with error code %(error)s' %
                                    {'session': session_id,
                                     'error': resp.status_code}))


def elcm_session_delete(irmc_info, session_id, terminate=False):
    """send an eLCM request to remove a session from the session list

    :param irmc_info: node info
    :param session_id: session id
    :param terminate: a running session must be terminated before removing
    :raises: ELCMSessionNotFound if the session does not exist
    :raises: SCCIClientError if SCCI failed
    """
    # Terminate the session first if needs to
    if terminate:
        # Get session status to check
        session = elcm_session_get_status(irmc_info, session_id)
        status = session['Session']['Status']

        # Terminate session if it is activated or running
        if status == 'running' or status == 'activated':
            elcm_session_terminate(irmc_info, session_id)

    # Send DELETE request to the server
    resp = elcm_request(irmc_info,
                        method='DELETE',
                        path='/sessionInformation/%s/remove' % session_id)

    if resp.status_code == 200:
        return
    elif resp.status_code == 404:
        raise ELCMSessionNotFound('Session "%s" does not exist' % session_id)
    else:
        raise scci.SCCIClientError(('Failed to remove session '
                                    '"%(session)s" with error code %(error)s' %
                                    {'session': session_id,
                                     'error': resp.status_code}))


def _process_session_data(irmc_info, operation, session_id,
                          session_timeout=BIOS_CONFIG_SESSION_TIMEOUT):
    """process session for Bios config backup/restore or RAID config operation

    :param irmc_info: node info
    :param operation: one of 'BACKUP_BIOS', 'RESTORE_BIOS' or 'CONFIG_RAID'
    :param session_id: session id
    :param session_timeout: session timeout
    :return: a dict with following values:
        {
            'bios_config': <data in case of BACKUP/RESTORE_BIOS operation>,
            'warning': <warning message if there is>
        }
    or
        {
            'raid_config': <data of raid adapter profile>,
            'warning': <warning message if there is>
        }
    """
    session_expiration = time.time() + session_timeout

    while time.time() < session_expiration:
        # Get session status to check
        session = elcm_session_get_status(irmc_info=irmc_info,
                                          session_id=session_id)

        status = session['Session']['Status']
        if status == 'running' or status == 'activated':
            # Sleep a bit
            time.sleep(5)
        elif status == 'terminated regularly':
            result = {}

            if operation == 'BACKUP_BIOS':
                # Bios profile is created, get the data now
                result['bios_config'] = elcm_profile_get(
                    irmc_info=irmc_info,
                    profile_name=PROFILE_BIOS_CONFIG)
            elif operation == 'RESTORE_BIOS':
                # Bios config applied successfully
                pass
            elif operation == 'CONFIG_RAID':
                # Getting raid config
                result['raid_config'] = elcm_profile_get(irmc_info,
                                                         PROFILE_RAID_CONFIG)

            # Cleanup operation by deleting related session and profile.
            # In case of error, report it as warning instead of error.
            try:
                elcm_session_delete(irmc_info=irmc_info,
                                    session_id=session_id,
                                    terminate=True)
                if operation == 'CONFIG_RAID':
                    return result

                # FIXME: Currently, creating a profile will restart the
                # machine, which will cause an error during IPI installation,
                # so temporarily comment out the operation of deleting the
                # profile.
                # elcm_profile_delete(irmc_info=irmc_info,
                #                    profile_name=PROFILE_BIOS_CONFIG)

            except scci.SCCIError as e:
                result['warning'] = e

            return result
        else:
            # Error occurred, get session log to see what happened
            session_log = elcm_session_get_log(irmc_info=irmc_info,
                                               session_id=session_id)

            raise scci.SCCIClientError(
                ('Failed to %(operation)s config. '
                 'Session log is "%(session_log)s".' %
                 {'operation': operation,
                  'session_log': jsonutils.dumps(session_log)}))

    else:
        raise ELCMSessionTimeout(
            ('Failed to %(operation)s config. '
             'Session %(session_id)s log is timeout.' %
             {'operation': operation,
              'session_id': session_id}))


def backup_bios_config(irmc_info):
    """backup current bios configuration

    This function sends a BACKUP BIOS request to the server. Then when the bios
    config data are ready for retrieving, it will return the data to the
    caller. Note that this operation may take time.

    :param irmc_info: node info
    :return: a dict with following values:
        {
            'bios_config': <bios config data>,
            'warning': <warning message if there is>
        }
    """
    result = {}

    # 1. Make sure there is no BiosConfig profile in the store
    try:
        # Get the profile first, if not found, then an exception
        # will be raised.
        result['bios_config'] = elcm_profile_get(
            irmc_info=irmc_info,
            profile_name=PROFILE_BIOS_CONFIG)
        # FIXME: Currently, creating a profile will restart the machine,
        # which will cause an error during IPI installation,
        # so temporarily comment out the operation of deleting the profile.
        # Profile found, delete it
        # elcm_profile_delete(irmc_info=irmc_info,
        #                    profile_name=PROFILE_BIOS_CONFIG)
    except ELCMProfileNotFound:
        # 2. Send request to create a new profile for BiosConfig
        session = elcm_profile_create(irmc_info=irmc_info,
                                      param_path=PARAM_PATH_BIOS_CONFIG)

        # 3. Profile creation is in progress, we monitor the session
        session_timeout = irmc_info.get('irmc_bios_session_timeout',
                                        BIOS_CONFIG_SESSION_TIMEOUT)
        return _process_session_data(
            irmc_info=irmc_info,
            operation='BACKUP_BIOS',
            session_id=session['Session']['Id'],
            session_timeout=session_timeout)

    return result


def restore_bios_config(irmc_info, bios_config):
    """restore bios configuration

    This function sends a RESTORE BIOS request to the server. Then when the
    bios
    is ready for restoring, it will apply the provided settings and return.
    Note that this operation may take time.

    :param irmc_info: node info
    :param bios_config: bios config
    """

    def _process_bios_config():
        try:
            if isinstance(bios_config, dict):
                input_data = bios_config
            else:
                input_data = jsonutils.loads(bios_config)

            # The input data must contain flag "@Processing":"execute" in the
            # equivalent section.
            bios_cfg = input_data['Server']['SystemConfig']['BiosConfig']
            bios_cfg['@Processing'] = 'execute'

            return input_data
        except (TypeError, ValueError, KeyError):
            raise scci.SCCIInvalidInputError(
                ('Invalid input bios config "%s".' % bios_config))

    # 1. Parse the bios config and create the input data
    input_data = _process_bios_config()

    # FIXME: Currently, creating a profile will restart the machine,
    # which will cause an error during IPI installation,
    # so temporarily comment out the operation of deleting the profile.
    # 2. Make sure there is no BiosConfig profile in the store
    # try:
    #     Get the profile first, if not found, then an exception
    #     will be raised.
    #    elcm_profile_get(irmc_info=irmc_info,
    #                     profile_name=PROFILE_BIOS_CONFIG)
    #     Profile found, delete it
    #    elcm_profile_delete(irmc_info=irmc_info,
    #                        profile_name=PROFILE_BIOS_CONFIG)
    # except ELCMProfileNotFound:
    #     Ignore this error as it's not an error in this case
    #    pass

    # 3. Send a request to apply the param values
    session = elcm_profile_set(irmc_info=irmc_info,
                               input_data=input_data)

    # 4. Param values applying is in progress, we monitor the session
    session_timeout = irmc_info.get('irmc_bios_session_timeout',
                                    BIOS_CONFIG_SESSION_TIMEOUT)
    _process_session_data(irmc_info=irmc_info,
                          operation='RESTORE_BIOS',
                          session_id=session['Session']['Id'],
                          session_timeout=session_timeout)


def get_secure_boot_mode(irmc_info):
    """Get the status if secure boot is enabled or not.

    :param irmc_info: node info.
    :raises: SecureBootConfigNotFound, if there is no configuration for secure
             boot mode in the bios.
    :return: True if secure boot mode is enabled on the node, False otherwise.
    """

    result = backup_bios_config(irmc_info=irmc_info)

    try:
        bioscfg = result['bios_config']['Server']['SystemConfig']['BiosConfig']
        return bioscfg['SecurityConfig']['SecureBootControlEnabled']

    except KeyError:
        msg = ("Failed to get secure boot mode from server %s. Upgrading iRMC "
               "firmware may resolve this issue." % irmc_info['irmc_address'])
        raise SecureBootConfigNotFound(msg)


def set_secure_boot_mode(irmc_info, enable):
    """Enable/Disable secure boot on the server.

    :param irmc_info: node info
    :param enable: True, if secure boot needs to be
                   enabled for next boot, else False.
    """

    bios_config_data = {
        'Server': {
            '@Version': '1.01',
            'SystemConfig': {
                'BiosConfig': {
                    '@Version': '1.01',
                    'SecurityConfig': {
                        'SecureBootControlEnabled': enable
                    }
                }
            }
        }
    }
    restore_bios_config(irmc_info=irmc_info, bios_config=bios_config_data)


def _update_raid_input_data(target_raid_config, raid_input):
    """Process raid input data.

    :param target_raid_config: node raid info
    :param raid_input: raid information for creating via eLCM
    :raises ELCMValueError: raise msg if wrong input
    :return: raid_input: raid input data which create raid configuration
        {
          "Server":{
            "HWConfigurationIrmc":{
              "@Processing":"execute",
              "Adapters":{
                "RAIDAdapter":[
                  {
                    "@AdapterId":"RAIDAdapter0",
                    "@ConfigurationType":"Addressing",
                    "LogicalDrives":{
                    "LogicalDrive":[
                        {
                          "@Number":0,
                          "@Action":"Create",
                          "RaidLevel":"1"
                        }
                      ]
                    }
                  }
                ]
               },
               "@Version":"1.00"
             },
             "@Version":"1.01"
           }
        }
    """

    logical_disk_list = target_raid_config['logical_disks']

    raid_input['Server']['HWConfigurationIrmc'].update({'@Processing':
                                                        'execute'})
    array_info = raid_input['Server']['HWConfigurationIrmc']['Adapters'][
        'RAIDAdapter'][0]
    array_info['LogicalDrives'] = {'LogicalDrive': []}
    array_info['Arrays'] = {'Array': []}

    for i, logical_disk in enumerate(logical_disk_list):
        physical_disks = logical_disk.get('physical_disks')

        # Auto create logical drive along with random physical disks.
        # Allow auto create along with raid 10 and raid 50
        # with specific physical drive.
        if not physical_disks or logical_disk['raid_level'] \
                in ('10', '50'):
            array_info['LogicalDrives']['LogicalDrive'].append(
                {'@Action': 'Create',
                 'RaidLevel': logical_disk['raid_level'],
                 'InitMode': 'fast'})
            array_info['LogicalDrives']['LogicalDrive'][i].update({
                "@Number": i})

        else:
            # Create array disks with specific physical servers
            arrays = {
                "@Number": i,
                "@ConfigurationType": "Setting",
                "PhysicalDiskRefs": {
                    "PhysicalDiskRef": []
                }
            }

            lo_drive = {
                "@Number": i,
                "@Action": "Create",
                "RaidLevel": "",
                "ArrayRefs": {
                    "ArrayRef": [
                    ]
                },
                "InitMode": "fast"
            }

            array_info['Arrays']['Array'].append(arrays)
            array_info['LogicalDrives']['LogicalDrive'].append(lo_drive)

            lo_drive.update({'RaidLevel': logical_disk['raid_level']})
            lo_drive['ArrayRefs']['ArrayRef'].append({"@Number": i})

            for element in logical_disk['physical_disks']:
                arrays['PhysicalDiskRefs']['PhysicalDiskRef'].append({
                    '@Number': element})
        if logical_disk['size_gb'] != "MAX":
            # Ensure correctly order these items in dict
            size = collections.OrderedDict()
            size['@Unit'] = 'GB'
            size['#text'] = logical_disk['size_gb']
            array_info['LogicalDrives']['LogicalDrive'][i]['Size'] = size

    return raid_input


def get_raid_adapter(irmc_info):
    """Collect raid information on the server.

    :param irmc_info: node info
    :returns: raid_adapter: get latest raid adapter information
    """

    # Update raid adapter, due to raid adapter cannot auto update after
    # created raid configuration.
    _create_raid_adapter_profile(irmc_info)

    return elcm_profile_get(irmc_info, PROFILE_RAID_CONFIG)


def _get_existing_logical_drives(raid_adapter):
    """Collect existing logical drives on the server.

    :param raid_adapter: raid adapter info
    :returns: existing_logical_drives: get logical drive on server
    """
    existing_logical_drives = []
    logical_drives = raid_adapter['Server']['HWConfigurationIrmc'][
        'Adapters']['RAIDAdapter'][0].get('LogicalDrives')
    if logical_drives is not None:
        for drive in logical_drives['LogicalDrive']:
            existing_logical_drives.append(drive['@Number'])

    return existing_logical_drives


def _create_raid_adapter_profile(irmc_info):
    """Attempt delete exist adapter then create new raid adapter on the server.

    :param irmc_info: node info
    :returns: result: a dict with following values:
                {
                    'raid_config': <data of raid adapter profile>,
                    'warning': <warning message if there is>
                }
    """

    try:
        # Attempt erase exist adapter on BM Server
        elcm_profile_delete(irmc_info, PROFILE_RAID_CONFIG)
    except ELCMProfileNotFound:
        # Ignore this error as it's not an error in this case
        pass

    session = elcm_profile_create(irmc_info, PARAM_PATH_RAID_CONFIG)

    # Monitoring currently session until done.
    session_timeout = irmc_info.get('irmc_raid_session_timeout',
                                    RAID_CONFIG_SESSION_TIMEOUT)

    return _process_session_data(irmc_info, 'CONFIG_RAID',
                                 session['Session']['Id'],
                                 session_timeout)


def create_raid_configuration(irmc_info, target_raid_config):
    """Process raid_input then perform raid configuration into server.

    :param irmc_info: node info
    :param target_raid_config: node raid information
    """

    if len(target_raid_config['logical_disks']) < 1:
        raise ELCMValueError(message="logical_disks must not be empty")

    # Check RAID config in the new RAID adapter. Must be erased before
    # create new RAID config.
    raid_adapter = get_raid_adapter(irmc_info)
    logical_drives = raid_adapter['Server']['HWConfigurationIrmc'][
        'Adapters']['RAIDAdapter'][0].get('LogicalDrives')
    session_timeout = irmc_info.get('irmc_raid_session_timeout',
                                    RAID_CONFIG_SESSION_TIMEOUT)
    if logical_drives is not None:
        # Delete exist logical drives in server.
        # NOTE(trungnv): Wait session complete and raise error if
        # delete raid config during FGI(Foreground Initialization) in-progress
        # in previous mechanism.
        delete_raid_configuration(irmc_info)
        # Updating raid adapter profile after deleted profile.
        raid_adapter = get_raid_adapter(irmc_info)

    # Create raid configuration based on target_raid_config of node
    raid_input = _update_raid_input_data(target_raid_config, raid_adapter)
    session = elcm_profile_set(irmc_info, raid_input)
    # Monitoring raid creation session until done.
    _process_session_data(irmc_info, 'CONFIG_RAID',
                          session['Session']['Id'],
                          session_timeout)


def delete_raid_configuration(irmc_info):
    """Delete whole raid configuration or one of logical drive on the server.

    :param irmc_info: node info
    """
    # Attempt to get raid configuration on BM Server
    raid_adapter = get_raid_adapter(irmc_info)
    existing_logical_drives = _get_existing_logical_drives(raid_adapter)
    # Ironic requires delete_configuration first. Will pass if blank raid
    # configuration in server.
    if not existing_logical_drives:
        return
    raid_adapter['Server']['HWConfigurationIrmc'].update({
        '@Processing': 'execute'})
    logical_drive = raid_adapter['Server']['HWConfigurationIrmc'][
        'Adapters']['RAIDAdapter'][0]['LogicalDrives']['LogicalDrive']

    for drive in logical_drive:
        drive['@Action'] = 'Delete'

    # Attempt to delete logical drive in the raid config
    session = elcm_profile_set(irmc_info, raid_adapter)
    # Monitoring raid config delete session until done.
    session_timeout = irmc_info.get('irmc_raid_session_timeout',
                                    RAID_CONFIG_SESSION_TIMEOUT)
    _process_session_data(irmc_info, 'CONFIG_RAID', session['Session']['Id'],
                          session_timeout)

    # Attempt to delete raid adapter
    elcm_profile_delete(irmc_info, PROFILE_RAID_CONFIG)


def set_bios_configuration(irmc_info, settings):
    """Set BIOS configurations on the server.

    :param irmc_info: node info
    :param settings: Dictionary containing the BIOS configuration.
    :raise: BiosConfigNotFound, if there is wrong settings for bios
    configuration.
    """

    bios_config_data = {
        'Server': {
            'SystemConfig': {
                'BiosConfig': {}
            }
        }
    }

    versions = elcm_profile_get_versions(irmc_info)
    server_version = versions['Server'].get('@Version')
    bios_version = \
        versions['Server']['SystemConfig']['BiosConfig'].get('@Version')

    if server_version:
        bios_config_data['Server']['@Version'] = server_version
    if bios_version:
        (bios_config_data['Server']['SystemConfig']['BiosConfig']
            ['@Version']) = bios_version

    configs = {}
    for setting_param in settings:
        setting_name = setting_param.get("name")
        setting_value = setting_param.get("value")
        # Revert-conversion from a string of True/False to boolean.
        # It will be raise failed if put "True" or "False" string value.
        if isinstance(setting_value, str):
            if setting_value.lower() == "true":
                setting_value = True
            elif setting_value.lower() == "false":
                setting_value = False
        try:
            type_config, config = BIOS_CONFIGURATION_DICTIONARY[
                setting_name].split("_")
            if type_config in configs.keys():
                configs[type_config][config] = setting_value
            else:
                configs.update({type_config: {config: setting_value}})
        except KeyError:
            raise BiosConfigNotFound("Invalid BIOS setting: %s"
                                     % setting_param)
    bios_config_data['Server']['SystemConfig']['BiosConfig'].update(configs)
    restore_bios_config(irmc_info, bios_config_data)


def get_bios_settings(irmc_info):
    """Get the current BIOS settings on the server

    :param irmc_info: node info.
    :returns: a list of dictionary BIOS settings
    """

    bios_config = backup_bios_config(irmc_info)['bios_config']
    bios_config_data = bios_config['Server']['SystemConfig']['BiosConfig']
    settings = []

    # TODO(trungnv): Allow working with multi levels of BIOS dictionary.
    for setting_param in BIOS_CONFIGURATION_DICTIONARY:
        type_config, config = BIOS_CONFIGURATION_DICTIONARY[
            setting_param].split("_")
        if config in bios_config_data.get(type_config, {}):
            value = str(bios_config_data[type_config][config])
            settings.append({'name': setting_param, 'value': value})
    return settings



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for IPMI Module.
"""

from unittest import mock

from pyghmi import exceptions as ipmi_exception
from pyghmi.ipmi import command as ipmi_command
import testtools

from scciclient.irmc import ipmi


@mock.patch.object(ipmi_command, 'Command', new=mock.Mock())
class IpmiTestCase(testtools.TestCase):
    """Tests for IPMI

    Unit Test Cases for getting information via ipmi raw command
    """

    def setUp(self):
        super(IpmiTestCase, self).setUp()

        self.info = {'irmc_address': "10.0.0.10",
                     'irmc_username': "admin",
                     'irmc_password': "admin",
                     }

    @mock.patch.object(ipmi, '_send_raw_command')
    def test_get_tpm_status_true(self, exec_mock):
        exec_mock.return_value = {'command': 0xF5, 'code': 0x00, 'netfn': 0x2F,
                                  'data': [0x80, 0x28, 0x00, 0xC0, 0xC0]}

        cmd = "0x2E 0xF5 0x80 0x28 0x00 0x81 0xC0"
        actual_out = ipmi.get_tpm_status(self.info)
        self.assertEqual(True, actual_out)
        exec_mock.assert_called_once_with(mock.ANY, cmd)

    @mock.patch.object(ipmi, '_send_raw_command')
    def test_get_tpm_status_false(self, exec_mock):
        exec_mock.return_value = {'command': 0xF5, 'code': 0x00, 'netfn': 0x2F,
                                  'data': [0x80, 0x28, 0x00, 0x80, 0x01]}
        cmd = "0x2E 0xF5 0x80 0x28 0x00 0x81 0xC0"

        actual_out = ipmi.get_tpm_status(self.info)
        self.assertEqual(False, actual_out)
        exec_mock.assert_called_once_with(mock.ANY, cmd)

    @mock.patch.object(ipmi, '_send_raw_command')
    def test_get_tpm_status_error_code(self, exec_mock):
        exec_mock.return_value = {'command': 0xF5, 'code': 0x01, 'netfn': 0x2F,
                                  'data': [0x80, 0x28, 0x00, 0x80, 0x01]}
        cmd = "0x2E 0xF5 0x80 0x28 0x00 0x81 0xC0"

        self.assertRaises(ipmi.IPMIFailure, ipmi.get_tpm_status, self.info)
        exec_mock.assert_called_once_with(mock.ANY, cmd)

    @mock.patch.object(ipmi, '_send_raw_command')
    def test_get_tpm_status_exception(self, exec_mock):
        exec_mock.side_effect = ipmi_exception.IpmiException

        cmd = "0x2E 0xF5 0x80 0x28 0x00 0x81 0xC0"

        self.assertRaises(ipmi.IPMIFailure, ipmi.get_tpm_status, self.info)
        exec_mock.assert_called_once_with(mock.ANY, cmd)

    @mock.patch.object(ipmi, '_send_raw_command')
    def test_get_pci(self, exec_mock):
        pci_device_ids = ['0x1000/0x0079', '0x2100/0x0080']

        exec_mock.side_effect = ({'command': 0xF1, 'code': 0x00, 'netfn': 0x2F,
                                  'data': [0x80, 0x28, 0x00, 0x00, 0x00, 0x05,
                                           0x00, 0x10, 0x79, 0x00, 0x34, 0x17,
                                           0x76, 0x11, 0x00, 0x04, 0x01]},
                                 {'command': 0xF1, 'code': 0xC9, 'netfn': 0x2F,
                                  'error': 'Parameter out of range',
                                  'data': [0x80, 0x28, 0x00]})

        cmd1 = "0x2E 0xF1 0x80 0x28 0x00 0x1A 0x1 0x00"
        cmd2 = "0x2E 0xF1 0x80 0x28 0x00 0x1A 0x2 0x00"
        actual_out = ipmi.get_pci_device(self.info, pci_device_ids)
        self.assertEqual(1, actual_out)
        exec_mock.assert_has_calls([mock.call(mock.ANY, cmd1),
                                    mock.call(mock.ANY, cmd2)])

    @mock.patch.object(ipmi, '_send_raw_command')
    def test_get_pci_blank(self, exec_mock):
        pci_device_ids = []

        actual_out = ipmi.get_pci_device(self.info, pci_device_ids)
        self.assertEqual(0, actual_out)
        self.assertTrue(exec_mock.called)

    @mock.patch.object(ipmi, '_send_raw_command')
    def test_get_pci_not_found(self, exec_mock):
        pci_device_ids = ['0x1111/0x1179', '0x2100/0x0080']

        exec_mock.side_effect = ({'command': 0xF1, 'code': 0x00, 'netfn': 0x2F,
                                  'data': [0x80, 0x28, 0x00, 0x00, 0x00, 0x05,
                                           0x00, 0x10, 0x79, 0x00, 0x34, 0x17,
                                           0x76, 0x11, 0x00, 0x04, 0x01]},
                                 {'command': 0xF1, 'code': 0xC9, 'netfn': 0x2F,
                                  'error': 'Parameter out of range',
                                  'data': [0x80, 0x28, 0x00]})
        cmd1 = "0x2E 0xF1 0x80 0x28 0x00 0x1A 0x1 0x00"
        cmd2 = "0x2E 0xF1 0x80 0x28 0x00 0x1A 0x2 0x00"
        actual_out = ipmi.get_pci_device(self.info, pci_device_ids)
        self.assertEqual(0, actual_out)
        exec_mock.assert_has_calls([mock.call(mock.ANY, cmd1),
                                    mock.call(mock.ANY, cmd2)])

    @mock.patch.object(ipmi, '_send_raw_command')
    def test_get_pci_exception(self, exec_mock):
        pci_device_ids = ['0x1111/0x1179', '0x2100/0x0080']

        exec_mock.side_effect = ipmi_exception.IpmiException('Error')

        cmd = "0x2E 0xF1 0x80 0x28 0x00 0x1A 0x1 0x00"

        e = self.assertRaises(ipmi.IPMIFailure,
                              ipmi.get_pci_device,
                              self.info,
                              pci_device_ids)
        exec_mock.assert_called_once_with(mock.ANY, cmd)
        self.assertEqual('IPMI operation \'GET PCI device quantity\' '
                         'failed: Error', str(e))



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import testtools

from scciclient.irmc import scci
from scciclient.irmc.viom import client as viom_client
from scciclient.irmc.viom import elcm as viom_elcm


class VIOMConfigurationTestCase(testtools.TestCase):

    def setUp(self):
        super(VIOMConfigurationTestCase, self).setUp()

        self.irmc_info = {
            'irmc_address': '10.124.196.159',
            'irmc_username': 'admin',
            'irmc_password': 'admin0',
            'irmc_port': 80,
            'irmc_auth_method': 'basic',
            'irmc_client_timeout': 60,
        }
        self.identification = 'viom_identification'
        self.configurator = viom_client.VIOMConfiguration(
            self.irmc_info, self.identification)

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def _test_terminate(self, mock_set, reboot=None):
        if reboot is None:
            self.configurator.terminate()
            reboot = False
        else:
            self.configurator.terminate(reboot=reboot)
        expected_json = {
            'VIOMManage': {
                'Manage': False,
                'Identification': 'viom_identification',
            },
            'InitBoot': reboot,
            'Mode': 'delete',
            '@Version': '1.00',
        }
        mock_set.assert_called_once_with(expected_json)

    def test_terminate(self):
        self._test_terminate()

    def test_terminate_reboot_true(self):
        self._test_terminate(reboot=True)

    def test_terminate_reboot_false(self):
        self._test_terminate(reboot=False)

    @staticmethod
    def _create_json_for_apply(slot_json, init_boot=False):
        return {
            'UseVirtualAddresses': True,
            'Mode': 'new',
            'InitBoot': init_boot,
            'VIOMManage': {
                'Manage': True,
                'Identification': 'viom_identification'
            },
            'Slots': {
                'Slot': [slot_json]
            },
            '@Version': '1.00',
        }

    @staticmethod
    def _create_json_before_apply(slot_json):
        return {
            'VIOMManage': {
                'Identification': 'viom_identification'
            },
            'Slots': {
                'Slot': [slot_json]
            },
            '@Version': '1.00',
        }

    @staticmethod
    def _create_json_slot_onbaord_lan(lan_port):
        lan_ports = []
        for port_idx in range(1, lan_port['@PortIdx']):
            lan_ports.append(
                {'@PortIdx': port_idx,
                 'BootPriority': 1,
                 'BootProtocol': 'None',
                 'PortEnable': True,
                 'UseVirtualAddresses': False})
        lan_ports.append(lan_port)
        return {
            '@SlotIdx': 0,
            'OnboardControllers': {
                'OnboardController': [
                    {'@OnboardControllerIdx': 1,
                     'LANAdapter': {
                         'Ports': {
                             'Port': lan_ports
                         }
                     }}
                ]
            }
        }

    @staticmethod
    def _create_json_slot_with_cna_lan(card_idx, port_idx, lan_function):
        return {
            '@SlotIdx': 0,
            'AddOnCards': {
                'AddOnCard': [
                    {'@AddOnCardIdx': card_idx,
                     'CNAAdapter': {
                         'Ports': {
                             'Port': [
                                 {'@PortIdx': port_idx,
                                  'PortEnable': True,
                                  'Functions': {
                                      'Function': [
                                          {'@FunctionIdx': 1,
                                           'LANFunction': lan_function},
                                      ]
                                  }}
                             ]
                         }
                     }}
                ]
            }
        }

    @staticmethod
    def _create_json_slot_with_cna_iscsi(card_idx, port_idx, iscsi_function,
                                         lan_function=None):
        if not lan_function:
            lan_function = {
                'FunctionEnable': False,
                'BootProtocol': 'None',
                'BootPriority': 1,
            }
        return {
            '@SlotIdx': 0,
            'AddOnCards': {
                'AddOnCard': [
                    {'@AddOnCardIdx': card_idx,
                     'CNAAdapter': {
                         'Ports': {
                             'Port': [
                                 {'@PortIdx': port_idx,
                                  'PortEnable': True,
                                  'Functions': {
                                      'Function': [
                                          {'@FunctionIdx': 1,
                                           'LANFunction': lan_function},
                                          {'@FunctionIdx': 3,
                                           'ISCSIFunction': iscsi_function}
                                      ]
                                  }}
                             ]
                         }
                     }}
                ]
            }
        }

    @staticmethod
    def _create_json_slot_with_fc(card_idx, fc_port):
        return {
            '@SlotIdx': 0,
            'AddOnCards': {
                'AddOnCard': [
                    {'@AddOnCardIdx': card_idx,
                     'FCAdapter': {
                         'Ports': {
                             'Port': [
                                 fc_port
                             ]
                         }
                     }}
                ]
            }
        }

    @staticmethod
    def _create_json_slot_with_cna_fcoe(card_idx, port_idx, fcoe_function,
                                        lan_function=None):
        if not lan_function:
            lan_function = {
                'FunctionEnable': False,
                'BootProtocol': 'None',
                'BootPriority': 1,
            }
        return {
            '@SlotIdx': 0,
            'AddOnCards': {
                'AddOnCard': [
                    {'@AddOnCardIdx': card_idx,
                     'CNAAdapter': {
                         'Ports': {
                             'Port': [
                                 {'@PortIdx': port_idx,
                                  'PortEnable': True,
                                  'Functions': {
                                      'Function': [
                                          {'@FunctionIdx': 1,
                                           'LANFunction': lan_function},
                                          {'@FunctionIdx': 2,
                                           'FCoEFunction': fcoe_function}
                                      ]
                                  }}
                             ]
                         }
                     }}
                ]
            }
        }

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_lan_port_to_onboard(self, mock_set):
        self.configurator.set_lan_port('LAN0-1')
        port = {
            '@PortIdx': 1,
            'PortEnable': True,
            'BootProtocol': 'None',
            'BootPriority': 1,
            'UseVirtualAddresses': False,
        }
        slot = VIOMConfigurationTestCase._create_json_slot_onbaord_lan(port)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_lan_port_to_onboard_with_virtual_mac(self, mock_set):
        self.configurator.set_lan_port('LAN0-9', mac='aa:bb:cc:dd:ee')
        port = {
            '@PortIdx': 9,
            'PortEnable': True,
            'BootProtocol': 'None',
            'BootPriority': 1,
            'UseVirtualAddresses': True,
            'VirtualAddress': {
                'MAC': 'aa:bb:cc:dd:ee'
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_onbaord_lan(port)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply(reboot=True)
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot,
                                                             init_boot=True))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_lan_port_to_cna(self, mock_set):
        self.configurator.set_lan_port('CNA5-6')
        function = {
            'FunctionEnable': True,
            'BootProtocol': 'None',
            'BootPriority': 1,
            'UseVirtualAddresses': False,
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_lan(
            5, 6, function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_lan_port_to_cna_with_virtual_mac(self, mock_set):
        self.configurator.set_lan_port('CNA9-1', mac='12:34:56:78:90')
        function = {
            'FunctionEnable': True,
            'BootProtocol': 'None',
            'BootPriority': 1,
            'UseVirtualAddresses': True,
            'VirtualAddress': {
                'MAC': '12:34:56:78:90'
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_lan(
            9, 1, function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_iscsi_volume_to_onboard_lan(self, mock_set):
        self.configurator.set_iscsi_volume(
            'LAN0-9',
            'iqn-2017-04.com.fujitsu:01',
            initiator_ip='192.168.11.11',
            initiator_netmask=24,
            target_iqn='iqn-2017-04.com.fujitsu:11',
            target_ip='192.168.22.22')
        port = {
            '@PortIdx': 9,
            'PortEnable': True,
            'BootProtocol': 'ISCSI',
            'BootPriority': 1,
            'ISCSIBootEnvironment': {
                'ISCSIInitiator': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:01',
                    'IPv4Address': '192.168.11.11',
                    'SubnetMask': '255.255.255.0',
                    'VLANId': 0,
                },
                'ISCSITarget': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:11',
                    'IPv4Address': '192.168.22.22',
                    'PortNumber': 3260,
                    'BootLUN': 0,
                    'AuthenticationMethod': 'None'
                }
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_onbaord_lan(port)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_iscsi_volume_to_cna_card(self, mock_set):
        self.configurator.set_iscsi_volume(
            'CNA1-2',
            'iqn-2017-04.com.fujitsu:01',
            initiator_ip='192.168.11.11',
            initiator_netmask=16,
            target_iqn='iqn-2017-04.com.fujitsu:11',
            target_ip='192.168.22.22')
        iscsi_function = {
            'FunctionEnable': True,
            'BootProtocol': 'ISCSI',
            'BootPriority': 1,
            'ISCSIBootEnvironment': {
                'ISCSIInitiator': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:01',
                    'IPv4Address': '192.168.11.11',
                    'SubnetMask': '255.255.0.0',
                    'VLANId': 0,
                },
                'ISCSITarget': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:11',
                    'IPv4Address': '192.168.22.22',
                    'PortNumber': 3260,
                    'BootLUN': 0,
                    'AuthenticationMethod': 'None'
                }
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_iscsi(
            1, 2, iscsi_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_iscsi_volume_to_cna_card_chap(self, mock_set):
        self.configurator.set_iscsi_volume(
            'CNA9-9',
            'iqn-2017-04.com.fujitsu:01',
            initiator_ip='192.168.11.11',
            initiator_netmask=30,
            target_iqn='iqn-2017-04.com.fujitsu:11',
            target_ip='192.168.22.22',
            chap_user='chap-user',
            chap_secret='chap-secret')
        iscsi_function = {
            'FunctionEnable': True,
            'BootProtocol': 'ISCSI',
            'BootPriority': 1,
            'ISCSIBootEnvironment': {
                'ISCSIInitiator': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:01',
                    'IPv4Address': '192.168.11.11',
                    'SubnetMask': '255.255.255.252',
                    'VLANId': 0,
                },
                'ISCSITarget': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:11',
                    'IPv4Address': '192.168.22.22',
                    'PortNumber': 3260,
                    'BootLUN': 0,
                    'AuthenticationMethod': 'CHAP',
                    'ChapUserName': 'chap-user',
                    'ChapSecret': 'chap-secret'
                }
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_iscsi(
            9, 9, iscsi_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_iscsi_volume_to_cna_card_mutual_chap(self, mock_set):
        self.configurator.set_iscsi_volume(
            'CNA2-1',
            'iqn-2017-04.com.fujitsu:01',
            initiator_ip='192.168.11.11',
            initiator_netmask=8,
            target_iqn='iqn-2017-04.com.fujitsu:11',
            target_ip='192.168.22.22',
            boot_prio=2,
            target_lun=3,
            chap_user='chap-user',
            chap_secret='chap-secret',
            mutual_chap_secret='mutual-secret')
        iscsi_function = {
            'FunctionEnable': True,
            'BootProtocol': 'ISCSI',
            'BootPriority': 2,
            'ISCSIBootEnvironment': {
                'ISCSIInitiator': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:01',
                    'IPv4Address': '192.168.11.11',
                    'SubnetMask': '255.0.0.0',
                    'VLANId': 0,
                },
                'ISCSITarget': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:11',
                    'IPv4Address': '192.168.22.22',
                    'PortNumber': 3260,
                    'BootLUN': 3,
                    'AuthenticationMethod': 'MutualCHAP',
                    'ChapUserName': 'chap-user',
                    'ChapSecret': 'chap-secret',
                    'MutualChapSecret': 'mutual-secret',
                }
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_iscsi(
            2, 1, iscsi_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_iscsi_volume_to_cna_card_initiator_dhcp(self, mock_set):
        self.configurator.set_iscsi_volume(
            'CNA1-2',
            'iqn-2017-04.com.fujitsu:01',
            initiator_dhcp=True,
            target_iqn='iqn-2017-04.com.fujitsu:11',
            target_ip='192.168.22.22')
        iscsi_function = {
            'FunctionEnable': True,
            'BootProtocol': 'ISCSI',
            'BootPriority': 1,
            'ISCSIBootEnvironment': {
                'ISCSIInitiator': {
                    'DHCPUsage': True,
                    'Name': 'iqn-2017-04.com.fujitsu:01',
                },
                'ISCSITarget': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:11',
                    'IPv4Address': '192.168.22.22',
                    'PortNumber': 3260,
                    'BootLUN': 0,
                    'AuthenticationMethod': 'None'
                }
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_iscsi(
            1, 2, iscsi_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_iscsi_volume_to_cna_card_target_dhcp(self, mock_set):
        self.configurator.set_iscsi_volume(
            'CNA1-2',
            'iqn-2017-04.com.fujitsu:01',
            initiator_ip='192.168.11.11',
            initiator_netmask=16,
            target_dhcp=True)
        iscsi_function = {
            'FunctionEnable': True,
            'BootProtocol': 'ISCSI',
            'BootPriority': 1,
            'ISCSIBootEnvironment': {
                'ISCSIInitiator': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:01',
                    'IPv4Address': '192.168.11.11',
                    'SubnetMask': '255.255.0.0',
                    'VLANId': 0,
                },
                'ISCSITarget': {
                    'DHCPUsage': True,
                    'AuthenticationMethod': 'None'
                }
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_iscsi(
            1, 2, iscsi_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_lan_and_iscsi_volume_to_cna_card(self, mock_set):
        self.configurator.set_lan_port('CNA1-2')
        self.configurator.set_iscsi_volume(
            'CNA1-2',
            'iqn-2017-04.com.fujitsu:01',
            initiator_ip='192.168.11.11',
            initiator_netmask=1,
            target_iqn='iqn-2017-04.com.fujitsu:11',
            target_ip='192.168.22.22')
        lan_function = {
            'FunctionEnable': True,
            'BootProtocol': 'None',
            'BootPriority': 1,
            'UseVirtualAddresses': False,
        }
        iscsi_function = {
            'FunctionEnable': True,
            'BootProtocol': 'ISCSI',
            'BootPriority': 1,
            'ISCSIBootEnvironment': {
                'ISCSIInitiator': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:01',
                    'IPv4Address': '192.168.11.11',
                    'SubnetMask': '128.0.0.0',
                    'VLANId': 0,
                },
                'ISCSITarget': {
                    'DHCPUsage': False,
                    'Name': 'iqn-2017-04.com.fujitsu:11',
                    'IPv4Address': '192.168.22.22',
                    'PortNumber': 3260,
                    'BootLUN': 0,
                    'AuthenticationMethod': 'None'
                }
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_iscsi(
            1, 2, iscsi_function, lan_function=lan_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_fc_volume_to_fc_card(self, mock_set):
        self.configurator.set_fc_volume('FC1-1', '11:22:33:44:55')
        port = {
            '@PortIdx': 1,
            'UseVirtualAddresses': False,
            'PortEnable': True,
            'BootProtocol': 'FC',
            'BootPriority': 1,
            'FCBootEnvironment': {
                'FCTargets': {
                    'FCTarget': [
                        {'@FCTargetIdx': 1,
                         'TargetWWPN': '11:22:33:44:55',
                         'TargetLUN': 0}
                    ]
                },
                'FCLinkSpeed': 'auto',
                'FCTopology': 'auto_loop',
                'SANBootEnable': True,
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_fc(1, port)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_fc_volume_to_fc_card_with_virtual_wwn(self, mock_set):
        self.configurator.set_fc_volume('FC2-1', '11:22:33:44:55',
                                        boot_prio=3, target_lun=2,
                                        initiator_wwnn='aa:bb:cc:dd:ee',
                                        initiator_wwpn='12:34:56:78:90')
        port = {
            '@PortIdx': 1,
            'UseVirtualAddresses': True,
            'VirtualAddress': {
                'WWNN': 'aa:bb:cc:dd:ee',
                'WWPN': '12:34:56:78:90'
            },
            'PortEnable': True,
            'BootProtocol': 'FC',
            'BootPriority': 3,
            'FCBootEnvironment': {
                'FCTargets': {
                    'FCTarget': [
                        {'@FCTargetIdx': 1,
                         'TargetWWPN': '11:22:33:44:55',
                         'TargetLUN': 2}
                    ]
                },
                'FCLinkSpeed': 'auto',
                'FCTopology': 'auto_loop',
                'SANBootEnable': True,
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_fc(2, port)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_fc_volume_to_cna_card(self, mock_set):
        self.configurator.set_fc_volume('CNA2-1', '11:22:33:44:55')
        fcoe_function = {
            'UseVirtualAddresses': False,
            'FunctionEnable': True,
            'BootProtocol': 'FC',
            'BootPriority': 1,
            'FCBootEnvironment': {
                'FCTargets': {
                    'FCTarget': [
                        {'@FCTargetIdx': 1,
                         'TargetWWPN': '11:22:33:44:55',
                         'TargetLUN': 0}
                    ]
                },
                'FCLinkSpeed': 'auto',
                'FCTopology': 'auto_loop',
                'SANBootEnable': True,
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_fcoe(
            2, 1, fcoe_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_fc_volume_to_cna_card_with_virtual_wwn(self, mock_set):
        self.configurator.set_fc_volume('CNA9-9', '11:22:33:44:55',
                                        boot_prio=2, target_lun=3,
                                        initiator_wwnn='aa:bb:cc:dd:ee',
                                        initiator_wwpn='12:34:56:78:90')
        fcoe_function = {
            'UseVirtualAddresses': True,
            'VirtualAddress': {
                'WWNN': 'aa:bb:cc:dd:ee',
                'WWPN': '12:34:56:78:90'
            },
            'FunctionEnable': True,
            'BootProtocol': 'FC',
            'BootPriority': 2,
            'FCBootEnvironment': {
                'FCTargets': {
                    'FCTarget': [
                        {'@FCTargetIdx': 1,
                         'TargetWWPN': '11:22:33:44:55',
                         'TargetLUN': 3}
                    ]
                },
                'FCLinkSpeed': 'auto',
                'FCTopology': 'auto_loop',
                'SANBootEnable': True,
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_fcoe(
            9, 9, fcoe_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    @mock.patch.object(viom_elcm.ELCMVIOMClient, 'set_profile')
    def test_set_lan_and_fc_volume_to_cna_card(self, mock_set):
        self.configurator.set_lan_port('CNA2-1')
        self.configurator.set_fc_volume('CNA2-1', '11:22:33:44:55')
        lan_function = {
            'FunctionEnable': True,
            'BootProtocol': 'None',
            'BootPriority': 1,
            'UseVirtualAddresses': False,
        }
        fcoe_function = {
            'UseVirtualAddresses': False,
            'FunctionEnable': True,
            'BootProtocol': 'FC',
            'BootPriority': 1,
            'FCBootEnvironment': {
                'FCTargets': {
                    'FCTarget': [
                        {'@FCTargetIdx': 1,
                         'TargetWWPN': '11:22:33:44:55',
                         'TargetLUN': 0}
                    ]
                },
                'FCLinkSpeed': 'auto',
                'FCTopology': 'auto_loop',
                'SANBootEnable': True,
            }
        }
        slot = VIOMConfigurationTestCase._create_json_slot_with_cna_fcoe(
            2, 1, fcoe_function, lan_function=lan_function)
        self.assertEqual(
            VIOMConfigurationTestCase._create_json_before_apply(slot),
            self.configurator.dump_json())
        self.configurator.apply()
        mock_set.assert_called_once_with(
            VIOMConfigurationTestCase._create_json_for_apply(slot))

    def test_set_lan_port_to_onboard_overwrite(self):
        self.configurator.set_lan_port('LAN0-9')
        self.test_set_lan_port_to_onboard_with_virtual_mac()

    def test_set_lan_port_to_cna_overwrite(self):
        self.configurator.set_lan_port('CNA5-6', mac='12:34:56:78:90')
        self.test_set_lan_port_to_cna()

    def test_set_iscsi_volume_to_onboard_lan_overwrite(self):
        self.configurator.set_iscsi_volume(
            'LAN0-9',
            'iqn-initiator',
            initiator_ip='192.168.99.99',
            initiator_netmask=32,
            target_iqn='iqn-target',
            target_ip='192.168.88.88')
        self.test_set_iscsi_volume_to_onboard_lan()

    def test_set_iscsi_volume_to_cna_card_overwrite(self):
        self.configurator.set_iscsi_volume(
            'CNA1-2',
            'iqn-initiator',
            initiator_ip='192.168.99.99',
            initiator_netmask=16,
            target_iqn='iqn-target',
            target_ip='192.168.88.88')
        self.test_set_iscsi_volume_to_cna_card()

    def test_set_fc_volume_to_fc_card_overwrite(self):
        self.configurator.set_fc_volume('FC2-1', '11:22:33:44:55')
        self.test_set_fc_volume_to_fc_card_with_virtual_wwn()

    def test_set_fc_volume_to_cna_card_overwrite(self):
        self.configurator.set_fc_volume('CNA2-1', '11:22:33:44:55',
                                        boot_prio=2, target_lun=3,
                                        initiator_wwnn='aa:bb:cc:dd:ee',
                                        initiator_wwpn='12:34:56:78:90')
        self.test_set_fc_volume_to_cna_card()


class PhysicalPortIDParseTestCase(testtools.TestCase):

    def _validate_handler(self, handler, handler_class, slot_type, card_type,
                          slot_idx, card_idx, port_idx):
        self.assertIsInstance(handler, handler_class)
        self.assertEqual(slot_type, handler.slot_type)
        self.assertEqual(card_type, handler.card_type)
        self.assertEqual(slot_idx, handler.slot_idx)
        self.assertEqual(card_idx, handler.card_idx)
        self.assertEqual(port_idx, handler.port_idx)

    def test_lan_onboard(self):
        handler = viom_client._parse_physical_port_id('LAN0-1')
        self._validate_handler(handler, viom_client._LANPortHandler,
                               viom_client.ONBOARD, viom_client.LAN,
                               0, 1, 1)

    def test_lan_onboard_lower(self):
        handler = viom_client._parse_physical_port_id('lan0-2')
        self._validate_handler(handler, viom_client._LANPortHandler,
                               viom_client.ONBOARD, viom_client.LAN,
                               0, 1, 2)

    def test_lan_onboard_cammel(self):
        handler = viom_client._parse_physical_port_id('Lan0-9')
        self._validate_handler(handler, viom_client._LANPortHandler,
                               viom_client.ONBOARD, viom_client.LAN,
                               0, 1, 9)

    def test_lan_addon(self):
        handler = viom_client._parse_physical_port_id('LAN1-3')
        self._validate_handler(handler, viom_client._LANPortHandler,
                               viom_client.ADDON, viom_client.LAN,
                               0, 1, 3)

    def test_fc_addon(self):
        handler = viom_client._parse_physical_port_id('FC2-5')
        self._validate_handler(handler, viom_client._FCPortHandler,
                               viom_client.ADDON, viom_client.FC,
                               0, 2, 5)

    def test_cna_addon(self):
        handler = viom_client._parse_physical_port_id('CNA9-2')
        self._validate_handler(handler, viom_client._CNAPortHandler,
                               viom_client.ADDON, viom_client.CNA,
                               0, 9, 2)

    def test_unkown_card(self):
        self.assertRaises(scci.SCCIInvalidInputError,
                          viom_client._parse_physical_port_id, 'HCA1-1')

    def test_slot_out_of_range(self):
        self.assertRaises(scci.SCCIInvalidInputError,
                          viom_client._parse_physical_port_id, 'CNA10-9')

    def test_port_out_of_range_min(self):
        self.assertRaises(scci.SCCIInvalidInputError,
                          viom_client._parse_physical_port_id, 'FC0-0')

    def test_port_out_of_range_max(self):
        self.assertRaises(scci.SCCIInvalidInputError,
                          viom_client._parse_physical_port_id, 'FC9-10')

    def test_public_validation(self):
        self.assertIsNone(viom_client.validate_physical_port_id('LAN0-2'))

    def test_public_validation_error(self):
        self.assertRaises(scci.SCCIInvalidInputError,
                          viom_client.validate_physical_port_id, 'CNA1-0')


class ConvertNetMaskTestCase(testtools.TestCase):
    def test_convert_zero(self):
        self.assertEqual('0.0.0.0',
                         viom_client._convert_netmask(0))

    def test_convert_max(self):
        self.assertEqual('255.255.255.255',
                         viom_client._convert_netmask(32))

    def test_convert_nagative(self):
        self.assertRaises(scci.SCCIInvalidInputError,
                          viom_client._convert_netmask, -1)

    def test_convert_too_large(self):
        self.assertRaises(scci.SCCIInvalidInputError,
                          viom_client._convert_netmask, 33)
